#define F_CPU 16000000UL 

#include <stdio.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <stdlib.h>            
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include "uart.h"
#include "songs.h"

// time is in units of 100ms
volatile long int maintime;
//time variable for different effects
volatile int time05;
volatile int time1;
volatile int time10;
volatile int time25;
volatile int time50;
volatile int time100;
volatile int time110;
volatile int time100_n;
volatile int time100_m;
volatile int time105;
volatile int time200;
volatile int time400;
volatile int time600;
volatile int time1000;
volatile int time3s;

volatile int change05 = 0;
volatile int change1 = 0;
volatile int change10 = 0;
volatile int change25 = 0;
volatile int change50 = 0;
volatile int change100 = 0;
volatile int change110 = 0;
volatile int change100_n = 0;
volatile int change105 = 0;
volatile int change200 = 0;
volatile int change400 = 0;
volatile int change600 = 0;
volatile int change1000 = 0;
volatile int change3s = 0;

//when stop_maintime is 1, stop the counter for maintime
volatile int stop_maintime = 0;
// when 3 seconds is up, 3s_signal becomes 1
volatile int three_signal = 1;
//for debouncing a button
volatile int button_value = 0;
volatile int ff = 0;
//for passing into functions that display regular effects
volatile int i_n = 0;
//for passing into functions that display local max effects
volatile int i_lm = 0;


//time count for changing display every 200ms 
#define t05  1
#define t1  2
#define t10  20
#define t25  50
#define t50  100
#define t100  200
#define t110  520
#define t100_n  200
#define t105  210
#define t200 400
#define t400 800
#define t600 1200
#define t3s  6000
//time count for changing display every 1s 
#define t1000 1000
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);


//**********************************************************
//timer 0 overflow ISR, interrupts every 1ms
ISR (TIMER0_COMPA_vect) { 
	//keeping track of all times for visual effect 
	//Decrement the time if not already zero
	if (time05>0)	//for .5ms
		--time05;
	else if (time05 == 0){
		change05 = 1;
		time05 = t05;
	}



	if (time1>0)//1 ms	
		--time1;
	else if (time1 == 0){
		change1 = 1;
		time1 = t1;
	}


	if (time10>0)	//10ms
		--time10;
	else if (time10 == 0){
		change10 = 1;
		time10 = t10;
	}

	if (time25>0)	//25ms
		--time25;
	else if (time25 == 0){
		change25 = 1;
		time25 = t25;
	}

	if (time50>0)	//50ms
		--time50;
	else if (time50 == 0){
		change50 = 1;
		time50 = t50;
	}

	if (time100_n>0)	//100n
		--time100_n;
	else if (time100_n == 0){
		change100_n = 1;
		time100_n = t100_n;
	}

	if (time100_m>0)	
		--time100_m;
	else if (time100_m == 0){
		time100_m = t100;
		if (stop_maintime == 0)
			maintime++;
	}

	if (time100>0)//100ms	
		--time100;
	else if (time100 == 0){
		change100 = 1;
		time100 = t100;
	}

	if (time105>0)	//105ms
		--time105;
	else if (time105 == 0){
		change105 = 1;
		time105 = t105;
	}

	if (time110>0)	//260ms
		--time110;
	else if (time110 == 0){
		change110 = 1;
		time110 = t110;
	}


	if (time200>0)	//200ms
		--time200;
	else if (time200 == 0){
		change200 = 1;
		time200 = t200;		
	}

	if (time400>0)	//400ms
		--time400;
	else if (time400 == 0){	
		change400 = 1;
		time400 = t400;
	}

	if (time600>0)	//600ms
		--time600;
	else if (time600 == 0){
		change600 = 1;
		time600 = t600;
	
	}

	if (time1000>0)	//1s
		--time1000;
	else if (time1000 == 0){
		change1000 = 1;
		time1000 = t1000;
		
	}

	if (time3s>0)	//3s
		--time3s;
	else if (time3s == 0)	{
		
		three_signal = 1;
		time3s = t3s;
	}

} 





/**
* Initialize the microcontroller.
*
* the time interval to change effects (different row/column) would be at like
* 0.2 seconds so need timer to be 
**/
void initialize(void)
{
	//set port A and C as output, connects to decoder
	//A connects to rows, C connects to cols
	//the last four bits are four bit select 
	// for example PORTA0 is lsb and PORTA3 is msb
	//Port D controls the motor and the Vdd, Gnd for the muxes
	//Port D0 controls the turning on of the system
	//Port D2 accelerates the effects by 100ms
	//assume that point (0,0) is the top left corner
	DDRA = 0xff;
	DDRC = 0xff;
	DDRD = 0x00;
	PORTA = 0x00;
	PORTC = 0x00;
	PORTD = 0xff;

	//set up timer 0 for timebase 
	TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR 
	OCR0A = 124;  		//set the compare re to 250 time ticks
	//set prescalar to divide by 64 
	TCCR0B= 3; //0b00001011;	
	// turn on clear-on-match
	TCCR0A= (1<<WGM01) ;


	sei();
}


//assume that vdd lines go straight down and ground is horizontal
//expand circle in to out effect 
void circleio(int i_n){
		int i_n2 = 4-(i_n)%5;
		PORTA = i_n2;
		//the outer borders
		for (int i = i_n2; i < 10 - i_n2;i++){
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = i;
			}
		}
		//all the rows that are not the first and last
		for (int i = i_n2; i < 10 - i_n2;i++){
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTA = i;
			}
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = i_n2;
			}
						change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = 9 - i_n2;
			}
		}
		PORTA = 9 - i_n2;
		for (int i = i_n2; i < 10 - i_n2;i++){
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = i; 
			}
		}
}

//expand circle out to in effect 
void circleoi(int i_n){
		int i_n2 = (i_n)%10;

		PORTA = i_n2;
		for (int i = i_n2; i < 10 - i_n2;i++){
			//cycle through
				PORTC = i;
			
		}
		//all the rows that are not the first and last
		for (int i = i_n2; i < 10 - i_n2;i++){	
				PORTA = i;
				PORTC = i_n2;
				PORTC = 9 - i_n2;
			
		}
		PORTA = 9 - i_n2;
		for (int i = i_n2; i < 10 - i_n2;i++){
		//cycle through
			PORTC = i;
			 
		}

}

//the left and right portions of the square when divided by the diagonal lines light up
void h_circleoi(int i_n){
		int i_n2 = (i_n)%10;
		PORTA = i_n2;
		for (int i = i_n2; i < 10 - i_n2;i++){
			change05 = 0;
			time05 = t1;	
			while (change05 == 0){
				PORTC = i;
			}
		}
		// the squares that are not the outmost one
		for (int i = i_n2; i < 10 - i_n2;i++){	
				PORTA = i;
				PORTC = i_n2;
				PORTC = 9 - i_n2;
			
		}
		PORTA = 9 - i_n2;
		for (int i = i_n2; i < 10 - i_n2;i++){
							change05 = 0;
			time05 = t1;	
			while (change05 == 0){
				PORTC = i;
			} 
		}

}

//the right portion of the square when divided by the diagonal lines light up
void q_circleoi(int i_n){
		int i_n2 = (i_n)%10;
		PORTA = i_n2;
		//outer border
		for (int i = i_n2; i < 10 - i_n2;i++){
			change05 = 0;
			time05 = t1;	
			while (change05 == 0){
				PORTC = i;
			}
		}
		//non outer border
		for (int i = i_n2; i < 10 - i_n2;i++){	
				PORTA = i;
				PORTC = i_n2;
				PORTC = 9 - i_n2;
			
		}
		PORTA = 9 - i_n2;
		//outer border
		for (int i = i_n2; i < 10 - i_n2;i++){
				PORTC = i; 
		}

}



//expand circle out to in to out effect 
void circleoio(int i_n){
		int i_n2 = (i_n)%10;
		PORTA = i_n2;
		//outer border
		for (int i = i_n2; i < 10 - i_n2;i++){
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = i;
			}
		}
		for (int i = i_n2; i < 10 - i_n2;i++){
			change1 = 0;
			time1 = t1;	
			//so that enough time to charge
			while (change1 == 0){
				PORTA = i;
			}
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = i_n2;
			}
						change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = 9 - i_n2;
			}
		}
		PORTA = 9 - i_n2;
		for (int i = i_n2; i < 10 - i_n2;i++){
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = i; 
			}
		}

}

//raindrop like (shining each row sequentially)
void raindrop(int i) {
	int i_n = 9 - (i/2) %10;
	//for one row
	PORTC = i_n;
	for (int j = 0; j < 10; j++){
	//cycle through column
		change1 = 0;
		time1 = t1;	
		while (change1 == 0){
		PORTA = j;
		}
	}
}


//cols light up one after another
void mv_v_lines(int i) {
	//to select the ground for an entire row
	int i_n = i/2;
	i_n = i_n %10;
	PORTA = i_n;
	for (int j = 0; j < 10; j++){
		change1 = 0;
		time1 = t1;	
		//wait to give enough time to charge
		//cycle through the rows
		while (change1 == 0){
		PORTC = j;
		}
	}
}

//simulation of fireworks
void firework_fast(){
	//first light up the middle 2 columns going up in 10ms intervals
	for (int i = 9; i > -1; i--){
		change10 = 0;
		time10 = t10;	
		//middle 2 columns
		while (change10 == 0){
			PORTA = 4;
			PORTC = i;
			PORTA = 5;
		}
	}
	int j =	0;
	//then sparkle each row 
	for (int i = 0; i < 10;i++){
		change10 = 0;
		time10 = t10;
		PORTC = i;
	//sparkle LED in rows not uniformly
		while (change10 == 0){		
			PORTA = j%10;
			PORTA = (j+3)%10;
			PORTA = (j+6)%10;
			PORTA = (j+9)%10;	
		}
		j++;
	}
}

//slower firework
void firework_slow(){
	//first light up the middle 2 columns going up in 10ms intervals
	for (int i = 0; i < 10; i++){
		change50 = 0;
		time50 = t50;	
		while (change50 == 0){
			PORTA = 4;
			PORTC = i;
			PORTA = 5;
		}
	}
	int j =	0;
	//then sparkle each row 
	for (int i = 9; i > -1;i--){
		change25 = 0;
		time25 = t25;
		PORTC = i;
		// to give the effect of randomess
		while (change25 == 0){		
			PORTA = j%10;
			PORTA = (j+3)%10;
			PORTA = (j+6)%10;
			PORTA = (j+9)%10;	
		}
		j++;
	}
}


//sparkle for 200ms
void sparkle() {
	change200 = 0;
	time200 = t200;
	int j = 0;
	//start the 200 ms
	while (change200 == 0){
			for (int i = 0; i < 20; i=i+3){
				change1 = 0;
				time1 = t1;	
				while (change1 == 0){
					PORTA = i%10;
					//for a little randomness in effect
					if (j%2 == 0)
					PORTC = (i + j)%10;
					else
					PORTC = (i + j + 1)%10;
				}			
			}
		j++;
		if (j == 10)
			j = 0;				
	}
	change200 = 0;
}

//diagonal lines going from top left to bottom right
void diagonalline(int i){
		//to select the ground for an entire row
	int i_n = i %20;
	for (int j = 0; j < i_n; j++){
		if (j < 10)
			PORTA = j;
		else
			PORTA = j - 10;
		change1 = 0;
		time1 = t1;	
		//for getting the diagonal lines
		PORTC = i_n - j + 10;
		while (change1 == 0){
		PORTC = i_n - j;
		}
	}
}

//appearing from corner and receding
void corner(int i){
	int i_n = i;
	i_n = i_n %20;
	if (i_n > 10)
		i_n = 20 - i_n;
	for (int k = 0; k < i_n; k++){
		PORTA = k;
		//all the LEDs in a corner
		for (int j = 0; j < i_n; j++){
			change1 = 0;
			time1 = t1;	
			while (change1 == 0){
				PORTC = j;
			}
		}
	}
} 


//turn all the leds on 2 times, each time 100 ms
void beat_full(){
	int count = 0;
	//two times
	while (count < 2){
		change100 = 0;
		time100 = t100;
		while (change100 == 0){
			for (int i = 0; i < 10; i++){
				PORTC = i;
				for (int j = 0;j < 10;j++){
					change05 = 0;
					time05 = 1;	
					//give LED enough time to charge
					while (change05 == 0){
						PORTA = j;
				}
			}		
		}
	}
		count++;
	}
}


void (*n_effects[9]) (int i) = {circleoio, circleoi,circleio,raindrop,mv_v_lines,corner,diagonalline,h_circleoi,q_circleoi};
//want local max to play to completion
void (*lm_effects[4]) (void) = {firework_fast,firework_slow,sparkle,beat_full};


//debounce a buttom every 200 ms
void debounce (){
	if ((PIND & 0x04) == 0){
		if (button_value == 1)
			ff = 1;
		else
			button_value = 1;
	}
	else
		button_value = 0;

}



int main(){
	//initialize timers and times
	initialize();
	time25 = t25;
	time50 = t50;
	time100 = t100;
	time200 = t200;
	time400 = t400;
	time600 = t600;	
	time1000 = t1000;
	time3s = t3s;
	uart_init();
	stdout = stdin = stderr = &uart_str;

	fprintf(stdout,"starting\n\r");
	//indexes for going through the 2D arrays
	volatile int mstd_index = 0;
	volatile int lm_index = 0;
	volatile int eff_i = 0;
	maintime = 0;
	while(maintime < duration){

		//load a different effect every 3 seconds
    	//use a switch to start or pause the program		
		if ((PIND & 0x01) == 0){
			stop_maintime = 1;
			}		
			else{
			if (stop_maintime == 1)
				stop_maintime = 0;
			if (change105 == 1){
				change105 = 0;
				//fast forward debouncing
				debounce();
				if (ff == 1){
				fprintf(stdout,"fast f\n\r");
					maintime = maintime + 1;
					ff = 0;
				}
			}

			if (three_signal == 1){
				three_signal = 0;
				//change i_n every 100 ms
				//don't want to fall too much behind
				while (((mstd_index + 1)*60 < maintime)&&(mstd_index < length_mstd - 1))
					mstd_index++;			
			    
				change110 = 0;
				//hash function
				int temp = (mstd[mstd_index][0] + mstd[mstd_index][1])*20;
				temp = temp%9;
				eff_i = temp;
				n_effects[eff_i](i_n);
				i_n = 0;
				fprintf(stdout,"eff_i %d n %ld %d\n\r",eff_i, maintime,mstd_index);
				mstd_index++;
			}
			if (change110 == 1){
			//change the index for selecting rows/columns in general effects
				i_n++;
				if (i_n == 40)
					i_n = 0;	
				change110 = 0;   
			}
			n_effects[eff_i](i_n);


			if (pulse[lm_index][0] <= maintime){
				//when miss a local max, just skip it

				while((pulse[lm_index][0] < maintime)&&(lm_index < length_pulse - 1)){
					lm_index++;
				}
				//choose effects according to magnitude
				if (pulse[lm_index][1] < 0.35){
						fprintf(stdout,"0 %ld %d\n\r",maintime,lm_index);
						lm_effects[0]();
				}	
				else if ((pulse[lm_index][1] > 0.35) && (pulse[lm_index][1] < 0.5)){
						fprintf(stdout,"1 %ld %d\n\r",maintime,lm_index);
						lm_effects[1]();
				}	
				else if ((pulse[lm_index][1] > 0.5) && (pulse[lm_index][1] < 0.75)){
						fprintf(stdout,"2 %ld %d\n\r",maintime,lm_index);
						lm_effects[2]();
				}	
				else {
						fprintf(stdout,"3 %ld %d\n\r",maintime,lm_index);
						lm_effects[3]();
				}						
				lm_index++;	
			}
		}

	}
	return 1;
}
